#
#  VoyageOfTheBeagleByDarwin is copyright 2005 Jeffrey G. Pittman.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#  MA 02111-1307 USA

# This is a Plone product version of Voyage of the Beagle, by Charles 
# Darwin, the text of which is available online in the public domain. 

# The original text from Project Gutenberg, for Voyage of the Beagle,
# has footnotes that restart numbering with each chapter, so that
# there are multiple [1] footnotes, for example.  This won't do for
# a single large structured text file, so this code renumbers the
# footnotes on a global basis, so that each one is unique. This
# code also produces individual chapter files, for which the handling
# of footnotes is simpler.

# The Project Gutenberg text for Voyage of the Beagle is from:
# http://www.gutenberg.org/dirs/etext97/vbgle11.txt. Some modifications
# were made to the original file:

# There were two cases of [1]), and [1];, which were changed to [1] ),
# and [1] ; manually so the footnote references would come out OK
# in the split() calls. Also, in the actual Voyage of the Beagle text,
# italics were marked by _italics_ but these were changed with a text
# editor to *italics* for use in structured text.

# See the original source for verification of copyright for your country
# or use. The pertinent information from the original source is present
# here in the file header.html.

from Products.Archetypes.public import BaseFolderSchema, Schema
from Products.Archetypes.public import StringField, IntegerField, FloatField, LinesField, BooleanField
from Products.Archetypes.public import LinesWidget, TextAreaWidget, IdWidget, IntegerWidget, DecimalWidget, StringWidget, BooleanWidget, MultiSelectionWidget
from Products.Archetypes.public import BaseFolder, registerType
from Products.VoyageOfTheBeagleByDarwin.config import PROJECTNAME
from Products.CMFCore import CMFCorePermissions

import os,sys,string

from Globals import package_home
from config import *
from AccessControl import ClassSecurityInfo, ModuleSecurityInfo
from reportlab.pdfgen.canvas import Canvas

PACKAGE_HOME = package_home(GLOBALS)

class VoyageOfTheBeagleByDarwinEngine:
    def __init__(self,
                 links_input):
        self.links_input = links_input

        # parse the links input
        self.links = []

        for cvs_input_line in self.links_input:
            items = eval('[%s]' % cvs_input_line) # we've validated by now, so good to go
            term_or_phrase, link = items

            if len(link) > 0:
                self.links.append((term_or_phrase, link))

        # open and split original text
        book_text = open(os.path.join(PACKAGE_HOME,'vbgle11.txt'),'r')
        # Doing the following, with .read(), hung.
        #book_text = open(os.path.join(PACKAGE_HOME,'vbgle11.txt'),'r').read()

        self.header_text, self.beagle_text = self.splitText(book_text)

    # from: http://www.onlamp.com/pub/a/python/2002/07/11/recipes.html
    def IsInt(self, str):
        """Is the given string an integer?"""
        try:int(str)
        except ValueError:return 0
        else:return 1
    
    def renumberFootnotes(self, text):
        # renumber the inline footnote references for use in the
        # single large stx file.
    
        footnote_count = 1
    
        new_text = []
    
        # there are no lines with more than one footnote in
        # Voyage of the Beagle, so that makes it simpler
        for line in text:
            if '[' in line:
                words = line.split()
                # look for [1]-type footnotes, but not at the beginning
                # of a line -- those are the actual footnote text entries,
                # which will be handled in a separate loop below.
                for word in words[1:]:
                    if word.startswith('[') and word.endswith(']'):
                        possible_number = word[1:-1]
                        if self.IsInt(possible_number):
                            footnote = "[%d]" % footnote_count
                            line = line.replace(word, footnote)
                            #print footnote
                            footnote_count += 1
                            break;
            new_text.append(line)
    
        # now renumber the footnote text entries
    
        footnote_count = 1
    
        new_text2 = []
    
        for line in new_text:
            if line.startswith('['):
                words = line.split()
                # now fix the actual footnote text entries
                if words[0].endswith(']'):
                    possible_number = words[0][1:-1]
                    if self.IsInt(possible_number):
                        # add the stx .. [ marker in front
                        words[0] = ".. [%d]" % footnote_count
                        #print words[0]
                        footnote_count += 1
                new_text2.append((' '.join(words)).rstrip() + '\n')
            else:
                new_text2.append(line.rstrip() + '\n')
    
        return new_text2
                        
    def stxifyFootnoteEntries(self, text):
        new_text = []
    
        for line in text:
            if line.startswith('['):
                words = line.split()
                if words[0].endswith(']'):
                    possible_number = words[0][1:-1]
                    if self.IsInt(possible_number):
                        # add the stx .. [ marker in front
                        words[0] = ".. %s" % words[0]
                new_text.append((' '.join(words)).rstrip() + '\n')
            else:
                new_text.append(line.rstrip() + '\n')
    
        return new_text
    
    def splitText(self, text):
        first_text = []
        second_text = []
        go = 0
        for line in text:
            if go:
                second_text.append(line)
            elif string.find(line,'PREFACE') >= 0:
                second_text.append(line)
                go = 1
            else:
                first_text.append(line)
    
        return (first_text, second_text)
    
    def stxifyParagraphs(self, text):
        # The Project Gutenberg text comes with line length
        # hard-coded, which makes searching for terms/phrases
        # that span line endings difficult. So, make a pass
        # looking for paragraph boundaries (blank lines) and
        # remove newlines within paragraphs.
    
        new_text = []
    
        paragraph = []
        pre_lines = [] # for handling lines needing original formatting
        for line in text:
            if len(line.strip()) == 0:
                new_text.append(line)
                # close the preceeding paragraph, removing newlines 
                # on all lines except last
                if len(paragraph) > 0:
                    new_text.append(' '.join([l.rstrip() for l in paragraph[:-1]] + [paragraph[-1]]))
                    # the following two lines kept the lines separate
                    #new_text = new_text + [l.rstrip() + ' ' for l in paragraph[:-1]]
                    #new_text.append(paragraph[-1])
                    paragraph = []
                if len(pre_lines) > 0:
                    new_text = new_text + pre_lines
                    pre_lines = []
            else:
                if not line.lstrip().startswith('-- ') and line.startswith(' ') or line.startswith('-----'):
                    if len(pre_lines) == 0:
                        pre_lines.append('\n')
                        pre_lines.append('::\n')
                        pre_lines.append('\n')
                        pre_lines.append('  ' + line)
                    else:
                        pre_lines.append('  ' + line)
                else:
                    if len(pre_lines) > 0:
                        pre_lines.append('  ' + line)
                    else:
                        paragraph.append(line)
    
        return new_text

    def singleQuoteBracketedPhrases(self, stxified_text):
        # takes in stxified text so that searching isn't
        # bothered by newlines.
        new_text = []
    
        for line in stxified_text:
            if not line.startswith('..') and string.find(line,'[') >= 0:
                words = line.split()
                new_line = []
                normal_words = []
                bracketed_phrase = []
                for word in words:
                    if len(bracketed_phrase) > 0:
                        if word.endswith(']'):
                            bracketed_phrase.append("%s\'" % word)
                            if len(normal_words) > 0:
                                new_line += normal_words
                            new_line += bracketed_phrase
                            normal_words = []
                            bracketed_phrase = []
                        else:
                            bracketed_phrase.append(word)
                    else:
                        if word.startswith('['):
                            if not word.endswith(']'):
                                bracketed_phrase.append("\'%s" % word)
                            else:
                                possible_number = word[1:-1]
                                if not self.IsInt(possible_number):
                                    bracketed_phrase.append("\'%s\'" % word)
                                    if len(normal_words) > 0:
                                        new_line += normal_words
                                    new_line += bracketed_phrase
                                    normal_words = []
                                    bracketed_phrase = []
                                else:
                                    # this is a footnote ref, e.g. [1]
                                    normal_words.append(word)
                        else:
                            normal_words.append(word)
                if len(bracketed_phrase) > 0:
                    if len(normal_words) > 0:
                        new_line += normal_words + bracketed_phrase
                        normal_words = []
                        bracketed_phrase = []
                if len(normal_words) > 0:
                    new_line += normal_words
                new_text.append(' '.join(new_line))
                new_text.append('\n')
            else:
                new_text.append(line)
    
        return new_text

    def moveFootnotesToBottom(self, text):
        new_text = []
    
        paragraph = []
        footnote_paragraphs = []
        for line in text:
            if len(line.strip()) == 0:
                new_text.append(line)
                if len(paragraph) > 0:
                    if paragraph[0].startswith('..'):
                        footnote_paragraphs.append(paragraph)
                    else:
                        new_text = new_text + paragraph
                    paragraph = []
            else:
                paragraph.append(line)
    
        for paragraph in footnote_paragraphs:
            new_text = new_text + ['\n'] + [line for line in paragraph]
    
        return new_text 
    
    def stxifyHeaderText(self, text):
        stxified_text = []
    
        stxified_text.append("This is the original header text::\n")
        stxified_text.append('\n')
        for line in text:
            stxified_text.append('  ' + line.rstrip() + '\n')
    
        return stxified_text
    
    def insertLinks(self, text):
        # rendered html will be structured text, so links are in
        # the form "term or phrase":http://whatever, just inserted
        # in the text input.

        for link in self.links:
            term_or_phrase = link[0]
            http_link = link[1]
            stx_link = "\"%s\":%s" % (term_or_phrase, http_link)
            text = text.replace(term_or_phrase, stx_link)

        return text

    def getStructuredTextForBook(self):
        stxified_header_text = self.stxifyHeaderText(self.header_text)
    
        beagle_text_1 = self.renumberFootnotes(self.beagle_text)
                        
        beagle_text_2 = self.stxifyParagraphs(beagle_text_1)
    
        beagle_text_3 = self.moveFootnotesToBottom(beagle_text_2)
    
        beagle_text_4 = self.singleQuoteBracketedPhrases(beagle_text_3)

        stxified_text = []
        for line in stxified_header_text + beagle_text_4:
            # to avoid confusion for definition lists in stx:
            # but don't do it on lines that start with ----...,
            # as these are part of the pre text tables of data
            # in Voyage of the Beagle.
            if not line.startswith('-----'):
                fixed_line = line.replace('--','-') 
            stxified_text.append(fixed_line)

        return self.insertLinks(''.join(stxified_text))
    
    def getStructuredTextForChapters(self):
        # get stx for the header, preface, and each chapter.
        
        stxified_footnote_entries_text = self.stxifyFootnoteEntries(self.beagle_text)
        
        stxified_text = self.stxifyParagraphs(stxified_footnote_entries_text)

        stxified_text_2 = self.singleQuoteBracketedPhrases(stxified_text)

        chapters = []

        current_chapter = []

        # first lines are added to preface
        for line in stxified_text_2:
            if string.find(line,'CHAPTER') >= 0:
                chapters.append(current_chapter)
                current_chapter = []
                current_chapter.append(line)
            elif string.find(line,'VOYAGE')  >= 0 or string.find(line,'End of Project') >= 0: 
                # (skipping title lines and the final end line)
                pass
            else:
                # to avoid confusion for definition lists in stx:
                # but don't do it on lines that start with ----...,
                # as these are part of the pre text tables of data
                # in Voyage of the Beagle.
                if not line.startswith('-----'):
                    fixed_line = line.replace('--','-') 
                current_chapter.append(fixed_line)
        chapters.append(current_chapter)

        final_chapters = []
        final_chapters.append(''.join(self.stxifyHeaderText(self.header_text)))
        for ch in chapters:
            final_chapters.append(self.insertLinks(''.join(ch)))
        return final_chapters
        # The following doesn't work here, but it does in a regular python test
        #return [''.join(self.stxifyHeaderText(self.header_text))] + [self.insertLinks(''.join(chapter)) for chapter in chapters]

voyage_of_the_beagle_schema = BaseFolderSchema +  Schema((
     StringField('id',
                 required=0, ## Still actually required, but
                             ## the widget will supply the missing value
                             ## on non-submits
                 mode="rw",
                 accessor="getId",
                 mutator="setId",
                 default=None,
                 widget=IdWidget(label="Short Name",
                                 description="Should not contain spaces, underscores or mixed case. "\
                                 "Short Name is part of the item's web address.",
                                 visible={'view' : 'visible'}),
                 ),
    StringField('title',
                required=1,
                searchable=1,
                default='',
                accessor='Title',
                widget=StringWidget(label="Title",
                                    description="This name shows in the navigation menu."),
                ),
    StringField('description',
                isMetadata=1,
                accessor='Description',
                widget=TextAreaWidget(label='Description', 
                              description='Description is optional, but not a bad idea for search text purposes.'),
                ),
    LinesField('book_links_field',
               required=0,
               default='',
               accessor="Links",
               validators=('validate_book_links',),
               widget=LinesWidget(label='Book Links',
                                  description='On each line, provide: \'term/phrase\', \'link\' (both in quotes, and separated by a comma), e.g., \'frog\',\'http://www.frogs.org\'',),
               ),
        ))

class VoyageOfTheBeagleByDarwin(BaseFolder):
    """
    A folderish object to hold Voyage of the Beagle
    """
    allowed_content_types = ['Document','Image',]

    filter_content_types = 1

    global_allow = 1

    schema = voyage_of_the_beagle_schema

    content_icon = 'voyageofthebeagleicon.gif'

    security = ClassSecurityInfo()

    actions = ({'id': 'view',
                'name': 'View',
                'action': 'string:${object_url}/VoyageOfTheBeagleByDarwin_view',
                'permissions': (CMFCorePermissions.View,)},
               {'id': 'references',
                'name': 'References',
                'action': 'string:${object_url}/reference_edit',
                'permissions': (CMFCorePermissions.ModifyPortalContent,),
                'visible':0},
               {'id': 'metadata',
                'name': 'Properties',
                'action': 'string:${object_url}/base_metadata',
                'permissions': (CMFCorePermissions.ModifyPortalContent,),
                'visible':0})

    def gatherInput(self):
        book_links_input = self.Links()

        return (book_links_input)
            
    # processForm is called for creation
    def processForm(self, data=1, metadata=0, REQUEST=None, values=None):
        ret = BaseFolder.processForm(self, data, metadata, REQUEST, values)

        self.createVoyageOfTheBeagleByDarwinEngine(self.gatherInput())

        return ret

    # update is called for edit
    def update(self, **kwargs):
        ret = BaseFolder.update(self, **kwargs)

        self.createVoyageOfTheBeagleByDarwinEngine(self.gatherInput())

        return ret

    def createVoyageOfTheBeagleByDarwinEngine(self, input_items):
        book_links_input = input_items

        self.engine = VoyageOfTheBeagleByDarwinEngine(book_links_input)
    
        existing_file = getattr(self, 'complete.html', None)
        if existing_file:
             self.manage_delObjects('complete.html')

        self.invokeFactory('Document', id='complete.html')
        html_file = self['complete.html']
        html_file.edit(text_format='structured-text',text=self.engine.getStructuredTextForBook())

        chapter_index = 0
        chapter_names = self.getChapterNames()
        for chapter_stx in self.engine.getStructuredTextForChapters():
            chapter_filename = self.getChapterFilename(chapter_names[chapter_index])
            existing_file = getattr(self, chapter_filename, None)
            if existing_file:
                self.manage_delObjects(chapter_filename)

            self.invokeFactory('Document', id=chapter_filename)
            html_file = self[chapter_filename]
            html_file.edit(text_format='structured-text',text=chapter_stx)
            chapter_index += 1

    def getChapterNames(self):
        return ['header', 'preface', 'chapter1', 'chapter2', 'chapter3', 'chapter4', 'chapter5', 'chapter6', 'chapter7', 'chapter8', 'chapter9', 'chapter10', 'chapter11', 'chapter12', 'chapter13', 'chapter14', 'chapter15', 'chapter16', 'chapter17', 'chapter18', 'chapter19', 'chapter20', 'chapter21']

    def getChapterFilenames(self):
        chapter_filenames = []
        for chapter in self.getChapterNames():
            chapter_filenames.append(self.getChapterFilename(chapter))
        
        return chapter_filenames

    def getFirstChapterFilename(self):
        first_chapter = self.getChapterNames()[0]
        return self.getChapterFilename(first_chapter)

    def getChapterFilename(self, chapter):
        chapter_filename = '%s.html' % chapter
        return chapter_filename

registerType(VoyageOfTheBeagleByDarwin, PROJECTNAME)
