import string

# To use this, make a directory and put in the original vbgle11.txt
# file along with this file. Then do python prepare_stx_files.py and
# a whole bunch of test stx files will be produced.

# The object of this code is to test file preparation functions in a 
# a Plone product version of Voyage of the Beagle, by Charles Darwin,
# the text of which is available online in the public domain.

# See the original online sources for verification of that for your 
# country or use.

# The original text from Project Gutenberg, for Voyage of the Beagle,
# has footnotes that restart numbering with each chapter, so that
# there are multiple [1] footnotes, for example.  This won't do for
# a single large structured text file, so this code renumbers the
# footnotes on a global basis, so that each one is unique. This
# code also produces individual chapter files, for which the handling
# of footnotes is simpler.

# The Project Gutenberg text for Voyage of the Beagle is from:
# http://www.gutenberg.org/dirs/etext97/vbgle11.txt.

# There were two cases of [1]), and [1];, which were changed to [1] ),
# and [1] ; manually so the footnote references would come out OK
# in the split() calls. Also, in the actual Voyage of the Beagle text,
# italics were marked by _italics_ but these were changed with a text
# editor to *italics* for use in structured text.

# from: http://www.onlamp.com/pub/a/python/2002/07/11/recipes.html
def IsInt(str):
    """Is the given string an integer?"""
    try:int(str)
    except ValueError:return 0
    else:return 1

def renumberFootnotes(text):
    # renumber the inline footnote references for use in the
    # single large file.

    footnote_count = 1

    new_text = []

    # there are no lines with more than one footnote in
    # Voyage of the Beagle, so that makes it simpler
    for line in text:
        if '[' in line:
            words = line.split()
            # look for [1]-type footnotes, but not at the beginning
            # of a line -- those are the actual footnote text entries,
            # which will be handled in a separate loop below.
            for word in words[1:]:
                if word.startswith('[') and word.endswith(']'):
                    possible_number = word[1:-1]
                    if IsInt(possible_number):
                        footnote = "[%d]" % footnote_count
                        line = line.replace(word, footnote)
                        print footnote
                        footnote_count += 1
                        break;
        new_text.append(line)

    # now renumber the footnote text entries

    footnote_count = 1

    new_text2 = []

    for line in new_text:
        if line.startswith('['):
            words = line.split()
            # now fix the actual footnote text entries
            if words[0].endswith(']'):
                possible_number = words[0][1:-1]
                if IsInt(possible_number):
                    # add the stx .. [ marker in front
                    words[0] = ".. [%d]" % footnote_count
                    print words[0]
                    footnote_count += 1
            new_text2.append((' '.join(words)).rstrip() + '\n')
        else:
            new_text2.append(line.rstrip() + '\n')

    return new_text2
                    
def stxifyFootnoteEntries(text):
    new_text = []

    for line in text:
        if line.startswith('['):
            words = line.split()
            if words[0].endswith(']'):
                possible_number = words[0][1:-1]
                if IsInt(possible_number):
                    # add the stx .. [ marker in front
                    words[0] = ".. %s" % words[0]
            new_text.append((' '.join(words)).rstrip() + '\n')
        else:
            new_text.append(line.rstrip() + '\n')

    return new_text

def singleQuoteBracketedPhrases(stxified_text):
    # takes in stxified text so that searching isn't
    # bothered by newlines.
    new_text = []

    for line in stxified_text:
        if not line.startswith('..') and string.find(line,'[') >= 0:
            words = line.split()
            new_line = []
            normal_words = []
            bracketed_phrase = []
            for word in words:
                if len(bracketed_phrase) > 0:
                    if word.endswith(']'):
                        bracketed_phrase.append("%s\'" % word)
                        if len(normal_words) > 0:
                            new_line += normal_words
                        new_line += bracketed_phrase
                        #print bracketed_phrase
                        normal_words = []
                        bracketed_phrase = []
                    else:
                        bracketed_phrase.append(word)
                else:
                    if word.startswith('['):
                        if not word.endswith(']'):
                            bracketed_phrase.append("\'%s" % word)
                        else:
                            possible_number = word[1:-1]
                            if not IsInt(possible_number):
                                bracketed_phrase.append("\'%s\'" % word)
                                if len(normal_words) > 0:
                                    new_line += normal_words
                                new_line += bracketed_phrase
                                #print bracketed_phrase
                                normal_words = []
                                bracketed_phrase = []
                            else:
                                # this is a footnote ref, e.g. [1]
                                normal_words.append(word)
                    else:
                        normal_words.append(word)
            if len(bracketed_phrase) > 0:
                if len(normal_words) > 0:
                    new_line += normal_words + bracketed_phrase
                    normal_words = []
                    bracketed_phrase = []
            if len(normal_words) > 0:
                new_line += normal_words
            new_text.append(' '.join(new_line))
            new_text.append('\n')
            #print
            #print ' '.join(new_line)
        else:
            new_text.append(line)

    return new_text

def splitText(text):
    first_text = []
    second_text = []
    go = 0
    for line in text:
        if go:
            second_text.append(line)
        elif string.find(line,'PREFACE') >= 0:
            second_text.append(line)
            go = 1
        else:
            first_text.append(line)

    return (first_text, second_text)

def stxifyParagraphs(text):
    # The Project Gutenberg text comes with line length
    # hard-coded, which makes searching for terms/phrases
    # that span line endings difficult. So, make a pass
    # looking for paragraph boundaries (blank lines) and
    # remove newlines within paragraphs, and the resulting
    # structured text will be fine paragraph-wise.

    new_text = []

    paragraph = []
    pre_lines = [] # for handling lines needing original formatting
    for line in text:
        if len(line.strip()) == 0:
            new_text.append(line)
            # close the preceeding paragraph, removing newlines 
            # on all lines except last
            if len(paragraph) > 0:
                new_text.append(' '.join([l.rstrip() for l in paragraph[:-1]] + [paragraph[-1]]))
                paragraph = []
            if len(pre_lines) > 0:
                new_text = new_text + pre_lines
                pre_lines = []
        else:
            if not line.lstrip().startswith('-- ') and line.startswith(' ') or line.startswith('-----'):
                if len(pre_lines) == 0:
                    pre_lines.append('\n')
                    pre_lines.append('::\n')
                    pre_lines.append('\n')
                    pre_lines.append('  ' + line)
                else:
                    pre_lines.append('  ' + line)
            else:
                if len(pre_lines) > 0:
                    pre_lines.append('  ' + line)
                else:
                    paragraph.append(line)

    return new_text

def moveFootnotesToBottom(text):
    new_text = []

    paragraph = []
    footnote_paragraphs = []
    for line in text:
        if len(line.strip()) == 0:
            new_text.append(line)
            if len(paragraph) > 0:
                if paragraph[0].startswith('..'):
                    footnote_paragraphs.append(paragraph)
                else:
                    new_text = new_text + paragraph
                paragraph = []
        else:
            paragraph.append(line)

    for paragraph in footnote_paragraphs:
        new_text = new_text + ['\n'] + [line for line in paragraph]

    return new_text 

def saveAsOneLargeFile(text, filename):
    # We can now save the whole book as one file.
    stxified_text = open(filename, 'w')
    for line in text:
        # to avoid confusion for definition lists in stx:
        # but don't do it on lines that start with ----...,
        # as these are part of the pre text tables of data
        # in Voyage of the Beagle.
        if not line.startswith('-----'):
            fixed_line = line.replace('--','-') 
    
        stxified_text.write(fixed_line)
    stxified_text.close()

def saveAsIndividualChapterFiles(text):
    # make a file for the preface, and each chapter.
    chapter_index = 0
    chapter_filenames = ['ch1.stx', 'ch2.stx', 'ch3.stx', 'ch4.stx', 'ch5.stx', 'ch6.stx', 'ch7.stx', 'ch8.stx', 'ch9.stx', 'ch10.stx', 'ch11.stx', 'ch12.stx', 'ch13.stx', 'ch14.stx', 'ch15.stx', 'ch16.stx', 'ch17.stx', 'ch18.stx', 'ch19.stx', 'ch20.stx', 'ch21.stx']
    current_chapter = open('preface.stx', 'w')
    for line in text:
        if string.find(line,'CHAPTER') >= 0:
            current_chapter.close()
            print chapter_index
            current_chapter = open(chapter_filenames[chapter_index], 'w')
            current_chapter.write(line)
            chapter_index += 1
        elif string.find(line,'VOYAGE')  >= 0 or string.find(line,'End of Project') >= 0: # skip title lines and the final end line
            pass
        else:
            # to avoid confusion for definition lists in stx:
            # but don't do it on lines that start with ----...,
            # as these are part of the pre text tables of data
            # in Voyage of the Beagle.
            if not line.startswith('-----'):
                fixed_line = line.replace('--','-') 
            current_chapter.write(fixed_line)
    current_chapter.close()

def stxifyHeaderText(text):
    stxified_text = []

    stxified_text.append("This is the original header text::\n")
    stxified_text.append('\n')
    for line in text:
        stxified_text.append('  ' + line.rstrip() + '\n')

    return stxified_text

def insertLinks(text):
    # rendered html will be structured text, so links are in
    # the form "term or phrase":http://whatever, just inserted
    # in the text input.

    links = []
    links.append(('bug','http://www.bug.com'))
    links.append(('gnat','http://www.gnat.com'))
    links.append(('insect','http://www.insect.com'))
    links.append(('frog','http://www.frog.com'))
    links.append(('mountain','http://www.mountain.com'))
    links.append(('lip','http://www.lip.com'))
    links.append(('river','http://www.river.com'))
    links.append(('smell','http://www.smell.com'))
    links.append(('beech','http://www.beech.com'))
    links.append(('Chilean species','http://www.butterfly.com'))
    links.append(('canoe','http://www.canoe.com'))

    for link in links:
        term_or_phrase = link[0]
        http_link = link[1]
        stx_link = "\"%s\":%s" % (term_or_phrase, http_link)
        text = text.replace(term_or_phrase, stx_link)

    return text

def getItAll(header_text, chapters):
    return [''.join(stxifyHeaderText(header_text))] + [insertLinks(''.join(chapter)) for chapter in chapters]

##############################
# open and split original text

original_text = open('vbgle11.txt', 'r')

header_text, beagle_text = splitText(original_text)
print len(header_text), len(beagle_text)

original_text.close()

#######################
# save a large stx file

stxified_header_text = stxifyHeaderText(header_text)

renumbered_footnotes_text = renumberFootnotes(beagle_text)
                    
stxified_and_renumbered_footnotes_text = stxifyParagraphs(renumbered_footnotes_text)

text_with_footnotes_at_bottom = moveFootnotesToBottom(stxified_and_renumbered_footnotes_text)

stxified_text_final = singleQuoteBracketedPhrases(text_with_footnotes_at_bottom)

saveAsOneLargeFile(stxified_header_text + stxified_text_final, 'vbgle11.stx')

###################################
# save individual chapter stx files

header_stx_file = open('header.stx', 'w')
for line in header_text:
    header_stx_file.write(line)
header_stx_file.close()

stxified_footnote_entries_text = stxifyFootnoteEntries(beagle_text)
print len(stxified_footnote_entries_text)

stxified_text = stxifyParagraphs(stxified_footnote_entries_text)

stxified_text_2 = singleQuoteBracketedPhrases(stxified_text)

chapters = []

current_chapter = []

# first lines are added to preface
for line in stxified_text_2:
    if string.find(line,'CHAPTER') >= 0:
        chapters.append(current_chapter)
        current_chapter = []
        current_chapter.append(line)
    elif string.find(line,'VOYAGE')  >= 0 or string.find(line,'End of Project') >= 0:
        # (skipping title lines and the final end line)
        pass
    else:
        # to avoid confusion for definition lists in stx:
        fixed_line = line.replace('--','-')
        current_chapter.append(fixed_line)
chapters.append(current_chapter)

index = 0
for something in getItAll(header_text, chapters):
    file = open("temp%d.stx" % index,'w')
    file.write(something)
    file.close()
    index += 1

